## cleaning the workspace
rm(list=ls(all=TRUE))

## loading packages
library(MortalitySmooth)
library(diffractometry)

## Construct of the gradient of the B-spline basis
bbgrad <- function(x, xl, xr, ndx, deg){
  dx <- (xr - xl) / ndx
  B0 <- MortSmooth_bbase(x,xl,xr,ndx,deg-1)/dx
  D <- diff(diag(ncol(B0) + 1))
  B <- B0 %*% D
  return(B)
}

## loading data 
data(indiumoxide)
## completed data
## angles
x0 <- indiumoxide[,1]
## photon-counts
y0 <- indiumoxide[,2]
## selecting data
whi <- which(x0>0 & x0<32.5)
x <- x0[whi]
y <- y0[whi]
## dimension
n <- length(x)
## log-counts
ly <- log(y)

## search the positions of the peaks
## by oversmoothing the data
## evaluating the first derivative
## and take where the derivative crosses zero

## large lambda with P-splines
fit0 <- Mort1Dsmooth(x=x, y=y,
                     method=3, lambda=10^6)
## fitted values
y0 <- fit0$fitted.val
## fitted linear predictor
ly0 <- log(fit0$fitted.val)
## B-splines bases for the derivatives
xl <- min(x)
xr <- max(x)
xmax <- xr + 0.01 * (xr - xl)
xmin <- xl - 0.01 * (xr - xl)
B1 <- bbgrad(x, xmin, xmax,
             ndx=fit0$ndx, deg=fit0$deg)
## first derivatives of the oversmooth values
y1 <- B1 %*% fit0$coef
## relative first derivatives:
## derivative of the linear predictor
ly1 <- y1/y0
## positions of the peaks
pos1 <- which(ly1[-1]<0 & ly1[1:(n-1)]>0)

## number of peaks
np <- length(pos1)

## finding the domains for each peaks
## by selecting the flexes around the peaks

## around the first peak
pos2.1 <- (pos1[1]-200):(pos1[1]+200)
x2.1 <- x[pos2.1]
p2 <- c(x2.1[which.max(ly1[pos2.1])],
        x2.1[which.min(ly1[pos2.1])])
## around the second peak
pos2.2 <- (pos1[2]-200):(pos1[2]+200)
x2.2 <- x[pos2.2]
p3 <- c(x2.2[which.max(ly1[pos2.2])],
        x2.2[which.min(ly1[pos2.2])])


## plotting
par(mfrow=c(2,1))
plot(x, ly1, t="l", col=2, lwd=2,
     ylab="first deriv. linear predictor")
abline(h=0)
abline(v=x[pos1], col=4, lwd=2)
abline(v=p2, col=3, lwd=2)
abline(v=p3, col=3, lwd=2)
plot(x, ly, t="l", ylab="linear predictor")
lines(x, ly0, t="l", col=2, lwd=2)
abline(v=x[pos1], col=4, lwd=2)
abline(v=p2, col=3, lwd=2)
abline(v=p3, col=3, lwd=2)
par(mfrow=c(1,1))


## weight for the first peak
x2 <- x[x>p2[1] & x<p2[2]]
w2 <- as.numeric(x%in%x2)
y2 <- y[which(w2==1)]
## weight for the second peak
x3 <- x[x>p3[1] & x<p3[2]]
w3 <- as.numeric(x%in%x3)
y3 <- y[which(w3==1)]
## weight for the baseline, for the starting values
x1 <- x[which(c(w2+w3)==0)]
w1 <- as.numeric(x%in%x1)


## constructing the bases

## B-splines for the baseline (over whole x-axis)
xl <- min(x)
xr <- max(x)
xmax <- xr + 0.01 * (xr - xl)
xmin <- xl - 0.01 * (xr - xl)
deg1 <- 80
B1 <- MortSmooth_bbase(x, xmin, xmax, deg1, 3)
nb1 <- ncol(B1)
## B-splines for the peaks
## first peak
deg2 <- floor(length(x2)/5)
x2l <- min(x2)
x2r <- max(x2)
x2max <- x2r + 0.01 * (x2r - x2l)
x2min <- x2l - 0.01 * (x2r - x2l)
B2 <- MortSmooth_bbase(x2, x2min, x2max, deg2, 3)
nb2 <- ncol(B2)
## second peak
deg3 <- floor(length(x3)/5)
x3l <- min(x3)
x3r <- max(x3)
x3max <- x3r + 0.01 * (x3r - x3l)
x3min <- x3l - 0.01 * (x3r - x3l)
B3 <- MortSmooth_bbase(x3, x3min, x3max, deg3, 3)
nb3 <- ncol(B3)

## complete model matrix as a list
## in their order and with dimensions
XX <- list(X1=B1, X2=B2, X3=B3)
nx <- length(XX)
nc <- unlist(lapply(XX, ncol))
## indicators for the coefficients of each basis
ind2 <- cumsum(nc)
ind1 <- c(1, ind2[1:(nx-1)]+1)
ind <- NULL
for(i in 1:nx){
  ind[[i]] <- ind1[i]:ind2[i]
}
## indicators for the fitted values
indF <- cbind(rep(1,n), w2, w3)


## starting values

## fit the baseline w/o considering
## the value around peaks
fit1 <- Mort1Dsmooth(x=x, y=y,
                     w=w1, 
                     method=3, lambda=10000,
                     ndx = deg1, deg = 3, pord = 2)
mu1.st <- exp(XX[[1]] %*% fit1$coef)
## fit the first peak, w/o baseline
y2.st <- y2
y2.st <- y2.st - mu1.st[w2==1]
y2.st[y2.st<=0] <- 0
fit2 <- Mort1Dsmooth(x=x2, y=y2.st,
                     method=3, lambda=100,
                     ndx = deg2, deg = 3, pord = 3)
## fit the second peak, w/o baseline
y3.st <- y3
y3.st <- y3.st - mu1.st[w3==1]
y3.st[y3.st<=0] <- 0
fit3 <- Mort1Dsmooth(x=x3, y=y3.st,
                     method=3, lambda=100,
                     ndx = deg3, deg = 3, pord = 3)

## starting values for both peaks
mu2.st <- exp(XX[[2]] %*% fit2$coef)
mu3.st <- exp(XX[[3]] %*% fit3$coef)
## and overall pattern
mu.st <- mu1.st
whi2 <- which(indF[,2]==1)
mu.st[whi2] <- mu.st[whi2] + mu2.st
whi3 <- which(indF[,3]==1)
mu.st[whi3] <- mu.st[whi3] + mu3.st


## plotting starting values
plot(x, y, t="l")
lines(x, mu.st, col=2, lwd=3)
lines(x, mu1.st, col=3, lwd=2, lty=3)
lines(x2, mu2.st, col=4, lwd=2, lty=3)
lines(x3, mu3.st, col=5, lwd=2, lty=3)

## concatenating the starting coefficients
coef.st <- as.vector(c(fit1$coef,
                       fit2$coef,
                       fit3$coef))

## penalty terms for the smoothing part
## difference matrices for the baseline
D1 <- diff(diag(nb1), diff=2)
tDD1 <- t(D1) %*% D1
## difference matrices for first peak
D2 <- diff(diag(nb2), diff=3)
tDD2 <- t(D2) %*% D2
## difference matrices for second peak
D3 <- diff(diag(nb3), diff=3)
tDD3 <- t(D3) %*% D3

## penalty for enforcing
## log-concaveness of the peaks
D2con <- diff(diag(nb2), diff=2)
w2con <- rep(0, nrow(D2con))
W2con <- diag(w2con)
D3con <- diff(diag(nb3), diff=2)
w3con <- rep(0, nrow(D3con))
W3con <- diag(w3con)
kappa <- 10^6

## smoothing parameters
lambdas1 <- 10^seq(4, 6, by=0.5)
nl1 <- length(lambdas1)
lambdas2 <- 10^seq(1, 4, by=0.5)
nl2 <- length(lambdas2)

## objects for saving outcomes
## for each lambdas combination
COEFS <- array(NA, dim=c(sum(nc), nl1, nl2),
             dimnames=list(NULL, lambdas1,lambdas2))
COEFS[,nl1,nl2] <- as.vector(coef.st)
BICs <- matrix(NA, nl1, nl2,
               dimnames=list(lambdas1, lambdas2))
ED1s <- matrix(NA, nl1, nl2,
               dimnames=list(lambdas1, lambdas2))
ED2s <- matrix(NA, nl1, nl2,
               dimnames=list(lambdas1, lambdas2))
ED3s <- matrix(NA, nl1, nl2,
               dimnames=list(lambdas1, lambdas2))
ITs <- matrix(NA, nl1, nl2,
              dimnames=list(lambdas1, lambdas2))
## solving step
d <- 0.2
## maximum number of iteration
max.it <- 100
## should we plot current estimations?
PLOT <- FALSE


## iterations for each smoothing parameters
## combinations
## this takes about 14 seconds on a 
## a 2.6 GHz portable PC with
## an Intel® Core™ i5-3320M processor
## and 4 GB of RAM

for(l2 in nl2:1){
  ## common lambdas for both peaks
  lambda2 <- lambdas2[l2]
  lambda3 <- lambda2
  ## penalty terms for the peaks
  P2 <- lambda2 * tDD2
  P3 <- lambda3 * tDD3
  for(l1 in nl1:1){
    ## lambda for the baseline
    lambda1 <- lambdas1[l1]
    ## penalty term for the baseline
    P1 <- lambda1 * tDD1
    ## final penalty term
    P <- bdiag(list(P1, P2, P3))
    ## "warm" starting values:
    ## previously estimated coefficients
    if(l1==nl1){
      if(l2==nl2) pos.st <- c(l1,l2)
      if(l2!=nl2) pos.st <- c(l1,l2+1)
    }else{
      pos.st <- c(l1+1,l2)
    }
    coef <- COEFS[,pos.st[1], pos.st[2]]

    ## plotting actual counts
    if(PLOT) plot(x, y, t="l", col=8)

    ## iterations for a given lambdas-combination
    for(it in 1:max.it){
      
      ## penalty terms for the shape constraints
      ## log-concaveness for the peaks
      P2con <- kappa * t(D2con)%*%W2con%*%D2con
      P3con <- kappa * t(D3con)%*%W3con%*%D3con
      Psha <- bdiag(list(diag(rep(0,nc[1])),
                         P2con, P3con))

      ## linear predictor
      eta <- numeric(nx*n)
      for(i in 1:nx){
        eta0 <- rep(0, n)
        etai <- XX[[i]]%*%coef[ind[[i]]]
        eta0[which(indF[,i]==1)]<- etai
        eta[1:n+(i-1)*n] <- eta0
      }
      ## components
      gamma <- exp(eta)*c(indF)
      ## expected values
      mu <- numeric(n)
      for(i in 1:nx){
        mu <- (1 * gamma[1:n+(i-1)*n]) + mu
      }
      if(PLOT){
        ## plotting overall fitted values
        lines(x, mu, col=2, lwd=3)
        ## plotting each component
        for(i in 1:nx){
          lines(x, gamma[1:n+(i-1)*n],
                col=i+2, lty=2)
        }
      }
      ## weights for the IWLS
      w <- mu
      ## modified model matrix for a CLM
      U <- matrix(NA, n, sum(nc))
      for(i in 1:nx){
        u <- gamma[1:n+(i-1)*n]/mu * 1
        XXi <- matrix(0, nrow=n, ncol=nc[i])
        XXi[which(indF[,i]==1), ] <- XX[[i]]
        U0 <- u * XXi
        U[,ind[[i]]] <- U0
      }
      ## regression parts for the P-CLM
      tUWU <- t(U) %*% (w * U)
      tUWUpP <- tUWU + P
      r <- y - mu
      tUr <- t(U) %*% r
      ## updating coefficients with a d-step 
      coef.old <- coef
      coef <- solve(tUWUpP, tUr + tUWU %*% coef)
      coef <- d*coef.old + (1-d)*coef      
      ## update weights for shape constraints
      W2con.old <- W2con
      coef2 <- coef[ind[[2]]]
      w2con <- diff(coef2, diff=2) >= 0
      W2con <- diag(w2con)
      W3con.old <- W3con
      coef3 <- coef[ind[[3]]]
      w3con <- diff(coef3, diff=2) >= 0
      W3con <- diag(w3con)
      ## convergence criterion for the coefficients
      dif.coef <- max(abs(coef.old-coef))/max(abs(coef))
      ## stopping loop at convergence
      if(dif.coef < 1e-04 & it > 4) break
    }
    ## Deviance
    sig <- sd(c(y-mu))
    dev <- 2*n*log(sig)
    ## effective dimensions
    H <- solve(tUWUpP, tUWU)
    h <- diag(H)
    ed1 <- sum(h[ind[[1]]])
    ed2 <- sum(h[ind[[2]]])
    ed3 <- sum(h[ind[[3]]])
    ## BIC
    bic <- dev + log(n)*(ed1+ed2+ed3)
    
    ## saving estimated values
    COEFS[,l1,l2] <- as.vector(coef)
    BICs[l1,l2] <- bic
    ED1s[l1,l2] <- ed1
    ED2s[l1,l2] <- ed2
    ED3s[l1,l2] <- ed3
    ITs[l1,l2] <- it
    ## monitoring part
    cat("", l1, "of", nl1, "\n",
            l2, "of", nl2, "\n",
            "iter: ", it, "\n",
            "dif:  ", dif.coef, "\n",
            "BIC:  ", bic, "\n",
        "\n", "\n")
  }
}

## selecting optimal smoothing parameters
pmin <- which(BICs==min(BICs), arr.ind=T)
lambda1.hat <- lambdas1[pmin[1]]
lambda2.hat <- lambdas2[pmin[2]]

## plotting BIC profile
labx <- expression(log[10](lambda^1))
laby <- expression(paste(log[10](lambda^2),
                         " , ", log[10](lambda^3)))
contour(log10(lambdas1), log10(lambdas2), BICs,
        xlab=labx, ylab=laby)
points(log10(lambda1.hat), log10(lambda2.hat),
       pch=3, col=2, lwd=3)

## fitted coefficients for the optimal lambdas
coef.hat <- COEFS[,pmin[1],pmin[2]]

## effective dimensions at the optimal lambdas
ed1.hat <- ED1s[pmin[1],pmin[2]]
ed2.hat <- ED2s[pmin[1],pmin[2]]
ed3.hat <- ED3s[pmin[1],pmin[2]]

## linear predictor for each component
etas <- NULL
for(i in 1:nx){
  etas[[i]] <- XX[[i]] %*% coef.hat[ind[[i]]]
}
eta1.hat <- etas[[1]]
eta2.hat <- etas[[2]]
eta3.hat <- etas[[3]]

## linear predictor in a matrix over the whole x
ETA.hat <- matrix(NA, n, nx)
for(i in 1:nx){
  ETA.hat[which(indF[,i]==1),i] <- etas[[i]]
}

## fitted values for each component
gamma1.hat <- exp(eta1.hat)
gamma2.hat <- exp(eta2.hat)
gamma3.hat <- exp(eta3.hat)
## expected values
mu.hat <- apply(exp(ETA.hat), 1, sum, na.rm=TRUE)

## deviance residuals
res1 <- sign(y - mu.hat)
res2 <- sqrt(2 * (y * log(y/mu.hat) - y + mu.hat))
res <- res1 * res2

## plotting fitted values
testofit <- c(expression(y),
              expression(mu),
              expression(gamma^1),
              expression(gamma^2),
              expression(gamma^3))
plot(x, y, col=8, t="l",
     xlab="angle",
     ylab="photon counts")
lines(x, mu.hat, col=2, lwd=2)
lines(x, gamma1.hat, col=3)
lines(x2, gamma2.hat, col=4)
lines(x3, gamma3.hat, col=5)
legend("top", legend=testofit, col=c(8,2:5), lty=1)
## plotting residuals
plot(x, res,
     xlab="angle",
     ylab="deviance residuals")
